/*
 * Diskheap defines...
 *
 * Copyright 2002 by Gray Watson
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted via http://256.com/gray/
 *
 * $Id: diskheap.h,v 1.27 2002/03/28 15:01:49 gray Exp $
 */

#ifndef __DISKHEAP_H__
#define __DISKHEAP_H__

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

/*
 * NOTE: positions in the diskheap are block-number and offset pairs
 * of unsigned integers.  Because of 32 bit limitations, a single
 * offset would only allow the diskheap to grow to 4gb.  With a
 * block-size of 512b, the block-number/offset pair can reference 2tb
 * and larger block sizes only increase this max (4096 == 16tb).
 *
 * The 0th block in the heap is _always_ where file header information
 * is stored.  This means at a 0 block-number is invalid and can be
 * used to mean end-of-file or other special things.  For instance, in
 * the diskheap_set_label function, if you pass in a block-number of
 * 0, it will clear a label entry.
 */

/*
 * Diskheap flags passed to the diskheap_open and diskheap_create
 * functions.
 *
 * Note: You cannot use the READ_ONLY flag with the diskheap_create
 * function.
 */
#define DISKHEAP_OPEN_READ_ONLY		(1<<0)	/* do not allow changes */
/* internal flags in diskheap_loc.h */

/*
 * Number of label to position reference points which can be stored in
 * the diskheap using the diskheap_label_set and diskheap_label_get
 * functions.
 *
 * WARNING: this number cannot be changed by the user and recompiled
 * since all of the user label information must fit into the first
 * block of the diskheap.
 */
#define DISKHEAP_NUM_LABELS		10

/*
 * error codes
 */
#define DISKHEAP_ERROR_NONE		1	/* no error */
#define DISKHEAP_ERROR_PNT		2	/* invalid diskheap_t pointer*/
#define DISKHEAP_ERROR_ARG_NULL		3	/* arg to function is NULL */
#define DISKHEAP_ERROR_ARG_INVALID	4	/* arg to function invalid */
#define DISKHEAP_ERROR_ALLOC		5	/* memory allocate failed */
#define DISKHEAP_ERROR_OPEN		6	/* opening file failed */
#define DISKHEAP_ERROR_READ		7	/* reading from file failed */
#define DISKHEAP_ERROR_WRITE		8	/* writing to file failed */
#define DISKHEAP_ERROR_LSEEK		9	/* lseek in file failed */
#define DISKHEAP_ERROR_FSYNC		10	/* fsyncing to file failed */
#define DISKHEAP_ERROR_VERSION		11	/* unknown disk version */
#define DISKHEAP_ERROR_INTERNAL		12	/* internal exception */
#define DISKHEAP_ERROR_READ_ONLY	13	/* file opened read-only */
#define DISKHEAP_ERROR_LABEL_EXISTS	14	/* label already exists */
#define DISKHEAP_ERROR_LABEL_LONG	15	/* label is too long */
#define DISKHEAP_ERROR_LABELS_FULL	16	/* label list is full */
#define DISKHEAP_ERROR_LABEL_UNKNOWN	17	/* label is unknown */
#define DISKHEAP_ERROR_LABEL_BOUNDS	18	/* label entry is oob */
#define DISKHEAP_ERROR_OFFSET_BOUNDS	19	/* offset is oob */
#define DISKHEAP_ERROR_OS_CAPABLE	20	/* OS cannot do that */
#define DISKHEAP_ERROR_DISK_MAGIC	21	/* invalid diskheap file */
#define DISKHEAP_ERROR_NO_FREE		22	/* no free space available */
#define DISKHEAP_ERROR_BLOCK_SIZE	23	/* invalid block size */
#define DISKHEAP_ERROR_BAD_CLOSE	24	/* heap not closed properly */
#define DISKHEAP_ERROR_AFTER_CREATE	25	/* heap was not just created */
#define DISKHEAP_ERROR_FREE_LIST	26	/* invalid internal free list*/
#define DISKHEAP_ERROR_BLOCK_HEADER	27	/* invalid block header */
#define DISKHEAP_ERROR_LOCATION		28	/* invalid block-num/offset */
#define DISKHEAP_ERROR_BLOCK_FREE	29	/* block has been freed */
#define DISKHEAP_ERROR_NOT_FOUND	30	/* block is not found */
#define DISKHEAP_ERROR_TOO_LARGE	31	/* allocation is too large */

/*
 * The following structure assists the caller in providing linear
 * access to the heap.  The block-num and offset combination can be
 * set with the diskheap_linear_first command or by looking up a label
 * or saving a previous allocation's location.
 *
 * With the diskheap_linear_first and next routines, you can step
 * through investigating all of the blocks in the heap.
 */
typedef struct {
  unsigned int		dl_block_num;	/* block number of current block */
  unsigned int		dl_offset;	/* offset of current block */
} diskheap_linear_t;

#ifdef __DISKHEAP_MAIN__

#include "diskheap_loc.h"

#else

/* generic table type */
typedef	void	diskheap_t;

#endif

/*<<<<<<<<<<  The below prototypes are auto-generated by fillproto */

/*
 * diskheap_t *diskheap_create
 *
 * DESCRIPTION:
 *
 * Create a new diskheap file.
 *
 * RETURNS:
 *
 * Success - A valid diskheap structure which should be passed to
 * diskheap_close.
 *
 * Failure - NULL
 *
 * ARGUMENTS:
 *
 * file -> File to open for the diskheap.
 *
 * flags -> Diskheap flags which affect the characteristics of the
 * heap.
 *
 * block_size -> Set the block_size of the diskheap.  Must be a factor
 * of 512bytes.  Set to 0 (recommended) to take the default of 4096
 * bytes.
 *
 * heap_type -> A type value which will be passed back to the
 * diskheap_open call.  This allows you to specify the type of data
 * being stored in the heap.
 *
 * open_mode -> Mode bits to pass to the open system call.  Set to 0
 * to take the default which is 0644.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
extern
diskheap_t	*diskheap_create(const char *file, const unsigned int flags,
				 const unsigned int block_size,
				 const unsigned int heap_type,
				 const unsigned int open_mode, int *error_p);

/*
 * int diskheap_set_free_space
 *
 * DESCRIPTION:
 *
 * Set the amount of disk space to reserve for free-space information.
 * The default is currently 1mb which should be able to store more
 * than 150,000 free slots in the file.  Each "free slot" represents a
 * block-number and size (in blocks) of a free area in the diskheap.
 * Contiguous free space is combined so each free area is bounded by
 * allocated areas.
 *
 * Free slots which cannot be accounted for in this area will not be
 * stored and may be lost however the default settings should make
 * this a rare occurrence.
 *
 * Note: this call has to be made _immediately_ following the call to
 * diskheap_create and no later.  Once the file has been created, it
 * cannot be adjusted.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are adjusting.
 *
 * free_space -> Number of bytes of free space to allow at the front
 * of the diskheap file.  The minimum is currently 512 bytes although
 * going this low will almost guarantee lost free locations.  It must
 * be a multiple of 512.
 */
extern
int	diskheap_set_free_space(diskheap_t *dh_p,
				const unsigned int free_space);

/*
 * diskheap_t *diskheap_open
 *
 * DESCRIPTION:
 *
 * Open a diskheap file.
 *
 * RETURNS:
 *
 * Success - A valid diskheap structure which should be passed to
 * diskheap_close.
 *
 * Failure - NULL
 *
 * ARGUMENTS:
 *
 * file -> File to open for the diskheap.
 *
 * flags -> Diskheap flags which affect the characteristics of the
 * heap.
 *
 * heap_type_p <- Pointer to an unsigned integer which, if not NULL,
 * will be set with the heap-type as given to diskheap_create.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
extern
diskheap_t	*diskheap_open(const char *file, const unsigned int flags,
			       unsigned int *heap_type_p, int *error_p);

/*
 * int diskheap_set_sync_often
 *
 * DESCRIPTION:
 *
 * Set to how often to sync the main administrative header and the
 * free list to the disk.
 *
 * Note: this may cause a sync to happen immediately.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are adjusting.
 *
 * sync_header -> How many store and delete transactions will cause
 * the admin header to be written to disk to update counters and other
 * stats.  Set to 0 to take the default which is every 100 trxn.  Set
 * to less than 0 for never in which case it will be updated when the
 * file is closed or when diskheap_sync_header() is called.
 *
 * sync_free -> How many store and delete transactions will cause the
 * free list to be written to disk to update which blocks are free and
 * which are in use.  Set to 0 to take the default which is every 100
 * trxn.  Set to less than 0 for never in which case it will be
 * updated when the file is closed or when diskheap_sync_free_list()
 * is called.
 */
extern
int	diskheap_set_sync_often(diskheap_t *dh_p,
				const int sync_header,
				const int sync_free);

/*
 * int diskheap_sync_header
 *
 * DESCRIPTION:
 *
 * Sync the header administrative information including counters,
 * transaction dates, and other statistics to disk.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are adjusting.
 */
extern
int	diskheap_sync_header(diskheap_t *dh_p);

/*
 * int diskheap_sync_free_list
 *
 * DESCRIPTION:
 *
 * Sync the free-list information to disk.  The free list manages
 * which blocks are free.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Diskheap pointer we are adjusting.
 */
extern
int	diskheap_sync_free_list(diskheap_t *dh_p);

/*
 * int diskheap_close
 *
 * DESCRIPTION:
 *
 * Close a diskheap structure and associated file.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to diskheap structure.
 */
extern
int	diskheap_close(diskheap_t *dh_p);

/*
 * int diskheap_store
 *
 * DESCRIPTION:
 *
 * Store a buffer of bytes into the diskheap returning the
 * block-number and offset where it was written.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * buffer -> Buffer of anonymous memory to be written to disk.
 *
 * user_size -> Size of buffer in bytes.  Must be > 0.
 *
 * user_type -> Type number to be associated with the buffer that is
 * being stored.  This number will be returned when you retrieve the
 * block and can be used to verify that you are dealing with the right
 * kind of storage chunk.  Pass in 0 if none is desired.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number of the allocation.  Both the block-number and the offset are
 * needed to read the information later.  This argument cannot be
 * NULL.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block of the allocation.  Both the block-number and
 * the offset are needed to read the information later.  This argument
 * cannot be NULL.
 */
extern
int	diskheap_store(diskheap_t *dh_p, const void *buffer,
		       const unsigned int user_size,
		       const unsigned int user_type,
		       unsigned int *block_num_p, unsigned int *offset_p);

/*
 * void *diskheap_retrieve
 *
 * DESCRIPTION:
 *
 * Read in a previously allocated block-number/offset/[size] from the
 * diskheap file.
 *
 * RETURNS:
 *
 * Success - Pointer to allocated memory holding the user's disk
 * information which must be passed to free to de-alloc.
 *
 * Failure - NULL
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * block_num -> Block-number of the allocation we are retrieving.
 *
 * offset -> Offset inside of the block of the allocation we are
 * retrieving.
 *
 * size_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the number of bytes in the total on-disk chunk, not
 * including any internal header information.
 *
 * type_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the type value which was passed to the store function.
 *
 * error_p <- Pointer to an integer which, if not NULL, will be set
 * with a diskheap error code.
 */
extern
void	*diskheap_retrieve(diskheap_t *dh_p, const unsigned int block_num,
			   const unsigned int offset, unsigned int *size_p,
			   unsigned int *type_p, int *error_p);

/*
 * int diskheap_retrieve_to_buf
 *
 * DESCRIPTION:
 *
 * Read in a previously allocated block-number/offset/[size] from the
 * diskheap file into the buffer provided.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * block_num -> Block-number of the allocation we are retrieving.
 *
 * offset -> Offset inside of the block of the allocation we are
 * retrieving.
 *
 * buffer -> Buffer to store the data in.
 *
 * max_read_size -> The maximum number of bytes to read from the data
 * item.  The buffer argument should be able to hold this number of
 * bytes.  Must be > 0.
 *
 * size_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the number of bytes in the total on-disk chunk, not
 * including any internal header information.
 *
 * type_p <- Pointer to an unsigned integer which, if not NULL, will
 * be set to the type value which was passed to the store function.
 */
extern
int	diskheap_retrieve_to_buf(diskheap_t *dh_p,
				 const unsigned int block_num,
				 const unsigned int offset, void *buffer,
				 const unsigned int max_read_size,
				 unsigned int *size_p, unsigned int *type_p);

/*
 * int diskheap_update
 *
 * DESCRIPTION:
 *
 * Update a previously allocated block with new data.  This calls
 * diskheap_delete and diskheap_store. If you wish to guarantee that a
 * program or system crash will not corrupt your data, you should set
 * the safer_b argument in which case it will do a diskheap_store
 * _first_, and then delete the old block.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * new_buffer -> Buffer of anonymous memory to be updated to disk.
 *
 * new_size -> Size of buffer in bytes.  Must be > 0.
 *
 * new_type -> Type number to be associated with the buffer that is
 * being written.  Pass in 0 if none is desired.
 *
 * old_block_num -> Block-number of the allocation we are updating.
 *
 * old_offset -> Offset inside of the block of the allocation we are
 * updating.
 *
 * safer_b -> Set to 1 to have this routine do the equivalent of a
 * diskheap_store _first_, and then the delete of the old block.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number of the allocation.  This argument cannot be NULL.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block of the allocation.  This argument cannot be
 * NULL.
 *
 * old_size_p <- Size of the user allocation that we updated.
 *
 * old_type_p <- Type of the user allocation that we updated.
 */
extern
int	diskheap_update(diskheap_t *dh_p, const void *new_buffer,
			const unsigned int new_size,
			const unsigned int new_type,
			const unsigned int old_block_num,
			const unsigned int old_offset, const int safer_b,
			unsigned int *block_num_p, unsigned int *offset_p,
			unsigned int *old_size_p, unsigned int *old_type_p);

/*
 * int diskheap_delete
 *
 * DESCRIPTION:
 *
 * Delete a previously stored block-number/offset section in the
 * diskheap file.  This puts it on the free list meaning that it can
 * be returned by the library at a later time.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * block_num -> Block-number of the allocation we are deleting.
 *
 * offset -> Offset inside of the block of the allocation we are
 * deleting.
 *
 * user_size_p <- Size of the user allocation in bytes.
 *
 * user_type_p <- Type number associated with the buffer by
 * diskheap_store.
 */
extern
int	diskheap_delete(diskheap_t *dh_p, const unsigned int block_num,
			const unsigned int offset, unsigned int *user_size_p,
			unsigned int *user_type_p);

/*
 * int diskheap_fsync
 *
 * DESCRIPTION:
 *
 * Call fsync on the file descriptor associated with the diskheap.  It
 * is designed to make sure that all buffered data gets moved to the
 * disk.  See the manual entry for fsync to determine what this does
 * in reality.
 *
 * Warning: even if the fsync succeeds, there is no guarantee that if
 * the system would immediately crash that the diskheap file would not
 * be corrupted.
 *
 * Warning: on some operating systems, fsync may not be available in
 * which case the DISKHEAP_ERROR_OS_CAPABLE error code will be
 * returned.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 */
extern
int	diskheap_fsync(diskheap_t *dh_p);

/*
 * int diskheap_label_set
 *
 * DESCRIPTION:
 *
 * Associate a specific location with a string label in the diskheap
 * header.  This allows you to establish the start of an on disk
 * structure so the diskheap file can be self contained with no
 * outside reference information.
 *
 * Examples of usage include the starting point for a hash table, the
 * first entry in a list, or the administrative information for a
 * mini-file system.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * label -> Label we are associated with a position.  The string must
 * be less than or equal to 32 bytes in length or an error will be
 * returned.  If the label matches an already set entry, then it will
 * be overwritten if the overwrite flag is set to 1 otherwise an error
 * will be returned.
 *
 * block_num -> Block number we are associating with the label.  If
 * set to 0 then delete the associated label.
 *
 * offset -> Offset number we are associating with the label.
 *
 * overwrite_b -> If set to 1 then allow the new position to overwrite
 * a current one if the label matches an existing entry.
 */
extern
int	diskheap_label_set(diskheap_t *dh_p, const char *label,
			   const unsigned int block_num,
			   const unsigned int offset,
			   const int overwrite_b);

/*
 * int diskheap_label_get
 *
 * DESCRIPTION:
 *
 * Get the location associated with a string label in the diskheap
 * header.  See diskheap_label_set for more information.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * label -> Label we are looking up in the header.  The string must be
 * less than or equal to 32 bytes in length or an error will be
 * returned.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number associated with the label.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block associated with the label.
 */
extern
int	diskheap_label_get(diskheap_t *dh_p, const char *label,
			   unsigned int *block_num_p,
			   unsigned int *offset_p);

/*
 * int diskheap_label_get_entry
 *
 * DESCRIPTION:
 *
 * There are a certain number (currently 10) of label and location
 * associations stored in the diskheap header.  This gets a specific
 * entry.  See diskheap_label_set for more information.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * entry_n -> Number of the label entry to get.
 *
 * label_p <- Pointer to a character pointer which will be set to an
 * allocated version of the label from the list.  This must be passed
 * to free to deallocate.
 *
 * block_num_p <- Pointer to an unsigned int which be set to the block
 * number associated with the label.
 *
 * offset_np <- Pointer to an unsigned int which be set to the offset
 * inside of the block associated with the label.
 */
extern
int	diskheap_label_get_entry(diskheap_t *dh_p, const unsigned int entry_n,
				 char **label_p, unsigned int *block_num_p,
				 unsigned int *offset_p);

/*
 * int diskheap_linear_first
 *
 * DESCRIPTION:
 *
 * Start the linear access function by setting the block-num and
 * offset to the first allocation in the heap.  If there are no
 * entries in the heap DISKHEAP_ERROR_NOT_FOUND will be returned.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * linear_p <-> Pointer to a linear access structure which will be
 * initialized to point to the first block in the heap.
 */
extern
int	diskheap_linear_first(diskheap_t *dh_p, diskheap_linear_t *linear_p);

/*
 * int diskheap_linear_next
 *
 * DESCRIPTION:
 *
 * Adjust the linear structure to reference the next allocation
 * location in the heap.  If there are no more entries in the heap
 * DISKHEAP_ERROR_NOT_FOUND will be returned.
 *
 * RETURNS:
 *
 * Success - DISKHEAP_ERROR_NONE
 *
 * Failure - DISKHEAP error code
 *
 * ARGUMENTS:
 *
 * dh_p <-> Pointer to a diskheap structure.
 *
 * linear_p <-> Pointer to a linear access structure which will be
 * adjusted to point to the first block in the heap.
 */
extern
int	diskheap_linear_next(diskheap_t *dh_p, diskheap_linear_t *linear_p);

/*
 * void diskheap_seed_random
 *
 * DESCRIPTION:
 *
 * This seeds the random number generator inside of the diskheap
 * library and sets a static flag which will not cause any more calls
 * to the random seed routine to be made.  This should be called
 * before any other diskheap calls are made to be effective.
 *
 * Note: this is usually only needed in testing since the library auto
 * seeds the random number generator the first time a diskheap is
 * opened or created.
 *
 * RETURNS:
 *
 * None.
 *
 * ARGUMENTS:
 *
 * seed -> Seed for the random number generator.  0 for none.
 */
extern
void	diskheap_seed_random(const int seed);

/*
 * const char *diskheap_strerror
 *
 * DESCRIPTION:
 *
 * Return string version of error codes.
 *
 * RETURNS:
 *
 * Returns the string equivalent of the error.
 *
 * ARGUMENTS:
 *
 * error -> the error value to lookup.
 */
extern
const char	*diskheap_strerror(const int error);

/*<<<<<<<<<<   This is end of the auto-generated output from fillproto. */

#ifdef __cplusplus
}
#endif

#endif /* ! __DISKHEAP_H__ */
