/*
 * Local defines for the diskheap module
 *
 * Copyright 2002 by Gray Watson.
 *
 * This file is part of the diskheap package.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose and without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies,
 * and that the name of Gray Watson not be used in advertising or
 * publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be reached via http://256.com/gray/
 *
 * $Id: diskheap_loc.h,v 1.25 2002/03/30 21:35:38 gray Exp $
 */

#ifndef __DISKHEAP_LOC_H__
#define __DISKHEAP_LOC_H__

#define SET_POINTER(pnt, val) \
	do { \
	  if ((pnt) != NULL) { \
	    (*(pnt)) = (val); \
          } \
        } while(0)

/* count of a statically allocated array */
#define COUNT_OF(array)		(sizeof(array) / sizeof(array[0]))

/*
 * bitflag tools for Variable and a Flag
 */
#define BIT_FLAG(x)		(1 << (x))
#define BIT_SET(v,f)		(v) |= (f)
#define BIT_CLEAR(v,f)		(v) &= ~(f)
#define BIT_IS_SET(v,f)		((v) & (f))
#define BIT_TOGGLE(v,f)		(v) ^= (f)

/* default block-size */
#define BLOCK_SIZE_FACTOR		512
#define DISKHEAP_DEFAULT_BLOCK_SIZE	4096

/*
 * Maximum level in the skip list.  This implies that we can only
 * store 2^32 entries optimally.  Needless to say this is plenty.
 */
#define DISKHEAP_MAX_SKIP_LEVEL		32

/*
 * Default number of store or delete trxn before the header or the
 * free information should be synced to disk.  See
 * diskheader_set_sync_often.
 */
#define DEFAULT_SYNC_HEADER		100
#define DEFAULT_SYNC_FREE		100

/*
 * Minimum amount of space in a sub-block including the sub-block
 * header information.  This will actually get whittled down to
 * a smaller minimum user allocation space.
 */
#define MINIMUM_SUB_BLOCK_SIZE		15

/*
 * structure which manager label -> position mappings for the user
 */
typedef struct {
  char		ll_label[32];	/* label for the position */
  unsigned int	ll_length;	/* length of the label string or 0 if none */
  unsigned int	ll_block_num;	/* block number of the location */
  unsigned int	ll_offset;	/* offset of the location */
} label_loc_t;

#define DISKHEAP_T_MAGIC	0xD15CD15C	/* dh_magic value */

/* external flags in diskheap.h */
#define DISKHEAP_JUST_CREATED		(1<<16)	/* file was just created */

/*
 * main diskheap structure
 */
typedef struct {
  unsigned int	dh_magic;	/* magic number to protect the structure */
  int		dh_fd;		/* file descriptor associated with diskheap */
  
  unsigned int	dh_flags;	/* flags defined in diskheap.h */
  unsigned int	dh_version;	/* diskheap version for backwards compat */
  unsigned int	dh_block_size;	/* size of the blocks in the file */
  unsigned int	dh_user_type;	/* type provided by the user */
  unsigned int	dh_min_block_n;	/* minimum block-number of user allocations */
  unsigned int	dh_max_block_n;	/* maximum block-number of user allocations */
  
  /* times of various operations */
  unsigned int	dh_create_time;	/* create time in epoch seconds */
  unsigned int	dh_open_time;	/* open time in epoch seconds */
  unsigned int	dh_close_time;	/* close time in epoch seconds */
  unsigned int	dh_sync_header_time; /* header last written */
  unsigned int	dh_sync_free_time; /* free-list last written */
  
  /* statistics which may not be valid depending on crash status */
  unsigned int	dh_alloc_c;	/* number of allocations in the file */
  unsigned int	dh_rewrite_c;	/* number of rewritten allocations */
  unsigned int	dh_delete_c;	/* number of deleted allocations */
  unsigned int	dh_max_alloc;	/* maximum allocation size */
  unsigned int	dh_min_alloc;	/* minimum allocation size */
  unsigned int	dh_total_blocks; /* total allocated _blocks_ */
  unsigned int	dh_invalid_free_c; /* invalid free slot counter */
  
  /* user string -> location areas */
  label_loc_t	dh_labels[DISKHEAP_NUM_LABELS];
  
  /*
   * free list information
   */
  /*
   * this is the root node in our skip-list where the entries are
   * sorted by block-number/offset position and then size
   */
  struct skip_slot_st	*dh_skip_pos_p;
  /*
   * this is the root node in our skip-list where the entries are
   * sorted by size and then block-number/offset postion
   */
  struct skip_slot_st	*dh_skip_size_p;
  /* node used to store the pointers when we are updating the entries */
  struct skip_slot_st	*dh_skip_update_p;
  
  /* sub-block information */
  unsigned int		dh_sub_sizes[30];	/* sizes of sub-block chunks */
  unsigned int		dh_sub_largest_size;	/* largest sub block alloc */
  
  /* working buffer */
  unsigned char	*dh_buf;		/* buffer of dh_block_size bytes */
  unsigned char	*dh_buf_bounds_p;	/* pnt past the end of block buffer */
  unsigned int	dh_max_header;		/* max block header size */
  unsigned int	dh_max_sub_header;	/* max sub-block header size */
  
  /* how often to sync */
  unsigned int	dh_sync_header;		/* # trxn before syncing header */
  unsigned int	dh_sync_header_c;	/* counter up to header sync */
  unsigned int	dh_sync_free;		/* # trxn before syncing free-list */
  unsigned int	dh_sync_free_list_c;	/* counter up to free sync */
  
} diskheap_t;

/* external diskheap structure for debuggers */
typedef diskheap_t	diskheap_ext_t;

/*
 * to map error to string
 */
typedef struct {
  int		es_error;		/* error number */
  char		*es_string;		/* associated string */
} error_str_t;

static	error_str_t	errors[]
#ifdef __GNUC__
__attribute__ ((unused))
#endif
= {
  { DISKHEAP_ERROR_NONE,		"no error" },
  { DISKHEAP_ERROR_PNT,			"invalid diskheap pointer" },
  { DISKHEAP_ERROR_ARG_NULL,		"argument is null" },
  { DISKHEAP_ERROR_ARG_INVALID,		"argument is invalid" },
  { DISKHEAP_ERROR_ALLOC,		"error allocating memory" },
  { DISKHEAP_ERROR_OPEN,		"open file failed" },
  { DISKHEAP_ERROR_READ,		"read from file failed" },
  { DISKHEAP_ERROR_WRITE,		"write to file failed" },
  { DISKHEAP_ERROR_LSEEK,		"lseek in file failed" },
  { DISKHEAP_ERROR_FSYNC,		"fsync of the diskheap file failed" },
  { DISKHEAP_ERROR_VERSION,		"unknown disk version number" },
  { DISKHEAP_ERROR_INTERNAL,		"internal exception failure" },
  { DISKHEAP_ERROR_READ_ONLY,		"diskheap was opened read only" },
  { DISKHEAP_ERROR_LABEL_EXISTS,	"label already exists" },
  { DISKHEAP_ERROR_LABEL_LONG,		"label is too long" },
  { DISKHEAP_ERROR_LABELS_FULL,		"label list is full" },
  { DISKHEAP_ERROR_LABEL_UNKNOWN,	"label is unknown" },
  { DISKHEAP_ERROR_LABEL_BOUNDS,	"label entry is out of bounds" },
  { DISKHEAP_ERROR_OFFSET_BOUNDS,	"file offset is out of bounds" },
  { DISKHEAP_ERROR_OS_CAPABLE,		"OS cannot perform function" },
  { DISKHEAP_ERROR_DISK_MAGIC,		"invalid diskheap file" },
  { DISKHEAP_ERROR_NO_FREE,		"no free space available in heap" },
  { DISKHEAP_ERROR_BLOCK_SIZE,		"invalid block size specified" },
  { DISKHEAP_ERROR_BAD_CLOSE,		"heap was not closed properly" },
  { DISKHEAP_ERROR_AFTER_CREATE,	"heap was not just created" },
  { DISKHEAP_ERROR_FREE_LIST,		"invalid internal free list" },
  { DISKHEAP_ERROR_BLOCK_HEADER,	"invalid on-disk block header" },
  { DISKHEAP_ERROR_LOCATION,		"invalid block-num/offset location" },
  { DISKHEAP_ERROR_BLOCK_FREE,		"block has been freed" },
  { DISKHEAP_ERROR_NOT_FOUND,		"block is not found" },
  { DISKHEAP_ERROR_TOO_LARGE,		"allocation is too large" },
  { 0, NULL }
};

#define INVALID_ERROR	"invalid error code"

#endif /* ! __DISKHEAP_LOC_H__ */
